@@ -3,73 +3,92 @@
 //! This is accomplished by finding the events that have happened since a certain
 //! time and inverting them.
 
-use std::convert::TryInto;
-use std::ffi::{OsStr, OsString};
+use std::ffi::OsString;
 use std::fmt::Write;
 use std::io::{stdin, BufRead, BufReader, Read};
 use std::sync::mpsc::{channel, Receiver, Sender, TryRecvError};
 use std::time::SystemTime;
 
 use cursive::event::Key;
+use cursive::traits::Resizable;
 use cursive::utils::markup::StyledString;
-use cursive::views::{Dialog, EditView, LinearLayout, OnEventView, ScrollView, TextView};
+use cursive::views::{Dialog, EditView, LinearLayout, OnEventView, Panel, ScrollView, TextView};
 use cursive::{Cursive, CursiveRunnable, CursiveRunner};
 use eyre::Context;
 use tracing::instrument;
 
-use crate::commands::smartlog::render_graph;
-use crate::core::eventlog::{Event, EventCursor, EventLogDb, EventReplayer, EventTransactionId};
-use crate::core::formatting::{printable_styled_string, Pluralize, StyledStringBuilder};
-use crate::core::graph::{make_graph, BranchOids, HeadOid, MainBranchOid};
-use crate::core::mergebase::MergeBaseDb;
-use crate::core::metadata::{
-    BranchesProvider, CommitMessageProvider, CommitOidProvider, DifferentialRevisionProvider,
-    HiddenExplanationProvider, RelativeTimeProvider,
-};
+use crate::commands::smartlog::{make_smartlog_graph, render_graph};
 use crate::declare_views;
-use crate::git::{CategorizedReferenceName, GitRunInfo, MaybeZeroOid, Repo};
-use crate::tui::{with_siv, Effects, SingletonView};
+use crate::tui::{with_siv, SingletonView};
+use lib::core::dag::Dag;
+use lib::core::effects::Effects;
+use lib::core::eventlog::{Event, EventCursor, EventLogDb, EventReplayer, EventTransactionId};
+use lib::core::formatting::{printable_styled_string, Glyphs, Pluralize, StyledStringBuilder};
+use lib::core::node_descriptors::{
+    BranchesDescriptor, CommitMessageDescriptor, CommitOidDescriptor,
+    DifferentialRevisionDescriptor, ObsolescenceExplanationDescriptor, Redactor,
+    RelativeTimeDescriptor,
+};
+use lib::git::{
+    check_out_commit, CategorizedReferenceName, CheckOutCommitOptions, GitRunInfo, MaybeZeroOid,
+    Repo, ResolvedReferenceInfo,
+};
 
 fn render_cursor_smartlog(
     effects: &Effects,
     repo: &Repo,
-    merge_base_db: &MergeBaseDb,
+    dag: &Dag,
     event_replayer: &EventReplayer,
     event_cursor: EventCursor,
 ) -> eyre::Result<Vec<StyledString>> {
-    let head_oid = event_replayer.get_cursor_head_oid(event_cursor);
-    let main_branch_oid = event_replayer.get_cursor_main_branch_oid(event_cursor, repo)?;
-    let branch_oid_to_names = event_replayer.get_cursor_branch_oid_to_names(event_cursor, repo)?;
-    let graph = make_graph(
+    let dag = dag.set_cursor(effects, repo, event_replayer, event_cursor)?;
+    let references_snapshot = event_replayer.get_references_snapshot(repo, event_cursor)?;
+
+    // FIXME: the information of which branch is checked out is not stored in
+    // the event log. For now, just pretend that `HEAD` was always detached in
+    // the past.
+    let head_info = ResolvedReferenceInfo {
+        oid: references_snapshot.head_oid,
+        reference_name: None,
+    };
+
+    let graph = make_smartlog_graph(
         effects,
         repo,
-        merge_base_db,
+        &dag,
         event_replayer,
         event_cursor,
-        &HeadOid(head_oid),
-        &MainBranchOid(main_branch_oid),
-        &BranchOids(branch_oid_to_names.keys().copied().collect()),
         true,
+        false,
     )?;
     let result = render_graph(
         effects,
         repo,
-        merge_base_db,
+        &dag,
         &graph,
-        &HeadOid(head_oid),
+        references_snapshot.head_oid,
         &mut [
-            &mut CommitOidProvider::new(true)?,
-            &mut RelativeTimeProvider::new(repo, SystemTime::now())?,
-            &mut HiddenExplanationProvider::new(&graph, event_replayer, event_cursor)?,
-            &mut BranchesProvider::new(repo, &branch_oid_to_names)?,
-            &mut DifferentialRevisionProvider::new(repo)?,
-            &mut CommitMessageProvider::new()?,
+            &mut CommitOidDescriptor::new(true)?,
+            &mut RelativeTimeDescriptor::new(repo, SystemTime::now())?,
+            &mut ObsolescenceExplanationDescriptor::new(event_replayer, event_cursor)?,
+            &mut BranchesDescriptor::new(
+                repo,
+                &head_info,
+                &references_snapshot,
+                &Redactor::Disabled,
+            )?,
+            &mut DifferentialRevisionDescriptor::new(repo, &Redactor::Disabled)?,
+            &mut CommitMessageDescriptor::new(&Redactor::Disabled)?,
         ],
     )?;
     Ok(result)
 }
 
-fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>> {
+fn describe_event(glyphs: &Glyphs, repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>> {
+    // Links to https://github.com/arxanas/git-branchless/issues/57
+    const EMPTY_EVENT_MESSAGE: &str =
+        "This may be an unsupported use-case; see https://git.io/J0b7z";
+
     let result = match event {
         Event::CommitEvent {
             timestamp: _,
@@ -79,13 +98,13 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Commit ")
-                    .append(repo.friendly_describe_commit_from_oid(*commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *commit_oid)?)
                     .build(),
                 StyledString::new(),
             ]
         }
 
-        Event::HideEvent {
+        Event::ObsoleteEvent {
             timestamp: _,
             event_tx_id: _,
             commit_oid,
@@ -99,13 +118,13 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Hide commit ")
-                    .append(repo.friendly_describe_commit_from_oid(*commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *commit_oid)?)
                     .build(),
                 StyledString::new(),
             ]
         }
 
-        Event::UnhideEvent {
+        Event::UnobsoleteEvent {
             timestamp: _,
             event_tx_id: _,
             commit_oid,
@@ -119,7 +138,7 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Unhide commit ")
-                    .append(repo.friendly_describe_commit_from_oid(*commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *commit_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -137,7 +156,7 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Check out to ")
-                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *new_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -154,11 +173,11 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Check out from ")
-                    .append(repo.friendly_describe_commit_from_oid(*old_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *old_oid)?)
                     .build(),
                 StyledStringBuilder::new()
                     .append_plain("            to ")
-                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *new_oid)?)
                     .build(),
             ]
         }
@@ -177,9 +196,7 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
                     .append_plain(CategorizedReferenceName::new(ref_name).render_full())
                     .build(),
                 StyledStringBuilder::new()
-                    .append_plain("This event should not appear. ")
-                    .append_plain("This is a (benign) bug -- ")
-                    .append_plain("please report it.")
+                    .append_plain(EMPTY_EVENT_MESSAGE)
                     .build(),
             ]
         }
@@ -197,7 +214,7 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
                     .append_plain("Create ")
                     .append_plain(CategorizedReferenceName::new(ref_name).friendly_describe())
                     .append_plain(" at ")
-                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *new_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -216,7 +233,7 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
                     .append_plain("Delete ")
                     .append_plain(CategorizedReferenceName::new(ref_name).friendly_describe())
                     .append_plain(" at ")
-                    .append(repo.friendly_describe_commit_from_oid(*old_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *old_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -236,13 +253,13 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
                     .append_plain("Move ")
                     .append_plain(ref_name.clone())
                     .append_plain(" from ")
-                    .append(repo.friendly_describe_commit_from_oid(*old_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *old_oid)?)
                     .build(),
                 StyledStringBuilder::new()
                     .append_plain("     ")
                     .append_plain(" ".repeat(ref_name.len()))
                     .append_plain("   to ")
-                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *new_oid)?)
                     .build(),
             ]
         }
@@ -256,11 +273,11 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Rewrite commit ")
-                    .append(repo.friendly_describe_commit_from_oid(*old_commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *old_commit_oid)?)
                     .build(),
                 StyledStringBuilder::new()
                     .append_plain("           as ")
-                    .append(repo.friendly_describe_commit_from_oid(*new_commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(glyphs, *new_commit_oid)?)
                     .build(),
             ]
         }
@@ -276,9 +293,7 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
                     .append_plain("Empty rewrite event.")
                     .build(),
                 StyledStringBuilder::new()
-                    .append_plain("This event should not appear. ")
-                    .append_plain("This is a (benign) bug -- ")
-                    .append_plain("please report it.")
+                    .append_plain(EMPTY_EVENT_MESSAGE)
                     .build(),
             ]
         }
@@ -287,13 +302,14 @@ fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>>
 }
 
 fn describe_events_numbered(
+    glyphs: &Glyphs,
     repo: &Repo,
     events: &[Event],
 ) -> Result<Vec<StyledString>, eyre::Error> {
     let mut lines = Vec::new();
     for (i, event) in (1..).zip(events) {
         let num_header = format!("{}. ", i);
-        for (j, event_line) in (0..).zip(describe_event(repo, event)?) {
+        for (j, event_line) in (0..).zip(describe_event(glyphs, repo, event)?) {
             let prefix = if j == 0 {
                 num_header.clone()
             } else {
@@ -315,7 +331,7 @@ fn select_past_event(
     mut siv: CursiveRunner<CursiveRunnable>,
     effects: &Effects,
     repo: &Repo,
-    merge_base_db: &MergeBaseDb,
+    dag: &Dag,
     event_replayer: &mut EventReplayer,
 ) -> eyre::Result<Option<EventCursor>> {
     #[derive(Clone, Copy, Debug)]
@@ -382,7 +398,7 @@ fn select_past_event(
                       event_cursor: EventCursor|
          -> eyre::Result<()> {
             let smartlog =
-                render_cursor_smartlog(effects, repo, merge_base_db, event_replayer, event_cursor)?;
+                render_cursor_smartlog(effects, repo, dag, event_replayer, event_cursor)?;
             SmartlogView::find(siv)
                 .get_inner_mut()
                 .set_content(StyledStringBuilder::from_lines(smartlog));
@@ -393,15 +409,13 @@ fn select_past_event(
                     "There are no previous available events.",
                 )],
                 Some((event_id, events)) => {
-                    let event_description = {
-                        let lines = describe_events_numbered(repo, events)?;
-                        StyledStringBuilder::from_lines(lines)
-                    };
-                    let relative_time_provider = RelativeTimeProvider::new(repo, now)?;
+                    let event_description_lines =
+                        describe_events_numbered(effects.get_glyphs(), repo, events)?;
+                    let relative_time_provider = RelativeTimeDescriptor::new(repo, now)?;
                     let relative_time = if relative_time_provider.is_enabled() {
                         format!(
                             " ({} ago)",
-                            RelativeTimeProvider::describe_time_delta(
+                            RelativeTimeDescriptor::describe_time_delta(
                                 now,
                                 events[0].get_timestamp()
                             )?
@@ -409,18 +423,18 @@ fn select_past_event(
                     } else {
                         String::new()
                     };
-                    vec![
-                        StyledStringBuilder::new()
-                            .append_plain("Repo after transaction ")
-                            .append_plain(events[0].get_event_tx_id().to_string())
-                            .append_plain(" (event ")
-                            .append_plain(event_id.to_string())
-                            .append_plain(")")
-                            .append_plain(relative_time)
-                            .append_plain(". Press 'h' for help, 'q' to quit.")
-                            .build(),
-                        event_description,
-                    ]
+
+                    let mut lines = vec![StyledStringBuilder::new()
+                        .append_plain("Repo after transaction ")
+                        .append_plain(events[0].get_event_tx_id().to_string())
+                        .append_plain(" (event ")
+                        .append_plain(event_id.to_string())
+                        .append_plain(")")
+                        .append_plain(relative_time)
+                        .append_plain(". Press 'h' for help, 'q' to quit.")
+                        .build()];
+                    lines.extend(event_description_lines);
+                    lines
                 }
             };
             InfoView::find(siv).set_content(StyledStringBuilder::from_lines(info_view_contents));
@@ -439,10 +453,15 @@ fn select_past_event(
             Ok(Message::Init) => {
                 let smartlog_view: SmartlogView = ScrollView::new(TextView::new("")).into();
                 let info_view: InfoView = TextView::new("").into();
-                siv.add_layer(
+                siv.add_fullscreen_layer(
                     LinearLayout::vertical()
-                        .child(smartlog_view)
-                        .child(info_view),
+                        .child(
+                            Panel::new(smartlog_view)
+                                .title("Commit graph")
+                                .full_height(),
+                        )
+                        .child(Panel::new(ScrollView::new(info_view)).title("Events"))
+                        .full_width(),
                 );
                 redraw(&mut siv, event_replayer, cursor)?;
             }
@@ -540,21 +559,21 @@ fn inverse_event(
             event_tx_id: _,
             commit_oid,
         }
-        | Event::UnhideEvent {
+        | Event::UnobsoleteEvent {
             timestamp: _,
             event_tx_id: _,
             commit_oid,
-        } => Event::HideEvent {
+        } => Event::ObsoleteEvent {
             timestamp,
             event_tx_id,
             commit_oid,
         },
 
-        Event::HideEvent {
+        Event::ObsoleteEvent {
             timestamp: _,
             event_tx_id: _,
             commit_oid,
-        } => Event::UnhideEvent {
+        } => Event::UnobsoleteEvent {
             timestamp,
             event_tx_id,
             commit_oid,
@@ -662,7 +681,7 @@ fn undo_events(
     }
 
     writeln!(effects.get_output_stream(), "Will apply these actions:")?;
-    let events = describe_events_numbered(repo, &inverse_events)?;
+    let events = describe_events_numbered(effects.get_glyphs(), repo, &inverse_events)?;
     for line in events {
         writeln!(
             effects.get_output_stream(),
@@ -689,12 +708,13 @@ fn undo_events(
     }
 
     let num_inverse_events = Pluralize {
-        amount: inverse_events.len().try_into().unwrap(),
-        singular: "inverse event",
-        plural: "inverse events",
+        determiner: None,
+        amount: inverse_events.len(),
+        unit: ("inverse event", "inverse events"),
     }
     .to_string();
 
+    let mut result = 0;
     for event in inverse_events.into_iter() {
         match event {
             Event::RefUpdateEvent {
@@ -710,17 +730,20 @@ fn undo_events(
                 // this case, rather than just update `HEAD` (and be left with a
                 // dirty working copy). The `Git` command will update the event
                 // log appropriately, as it will invoke our hooks.
-                git_run_info
-                    .run(
-                        effects,
-                        Some(event_tx_id),
-                        &[
-                            OsStr::new("checkout"),
-                            OsStr::new("--detach"),
-                            target_oid.as_os_str(),
-                        ],
-                    )
-                    .wrap_err_with(|| "Updating to previous HEAD location")?;
+                let exit_code = check_out_commit(
+                    effects,
+                    git_run_info,
+                    Some(event_tx_id),
+                    Some(target_oid.as_os_str()),
+                    &CheckOutCommitOptions {
+                        additional_args: &["--detach"],
+                        render_smartlog: true,
+                    },
+                )
+                .wrap_err("Updating to previous HEAD location")?;
+                if exit_code != 0 {
+                    result = exit_code;
+                }
             }
             Event::RefUpdateEvent {
                 timestamp: _,
@@ -741,9 +764,7 @@ fn undo_events(
                 message: _,
             } => match repo.find_reference(&ref_name)? {
                 Some(mut reference) => {
-                    reference
-                        .delete()
-                        .wrap_err_with(|| "Applying `RefUpdateEvent`")?;
+                    reference.delete().wrap_err("Applying `RefUpdateEvent`")?;
                 }
                 None => {
                     writeln!(
@@ -773,8 +794,8 @@ fn undo_events(
                 repo.create_reference(&ref_name, new_oid, true, "branchless undo")?;
             }
             Event::CommitEvent { .. }
-            | Event::HideEvent { .. }
-            | Event::UnhideEvent { .. }
+            | Event::ObsoleteEvent { .. }
+            | Event::UnobsoleteEvent { .. }
             | Event::RewriteEvent { .. } => {
                 event_log_db.add_events(vec![event])?;
             }
@@ -786,25 +807,46 @@ fn undo_events(
         "Applied {}.",
         num_inverse_events
     )?;
-    Ok(0)
+    Ok(result)
 }
 
 /// Restore the repository to a previous state interactively.
 #[instrument]
-pub fn undo(effects: &Effects, git_run_info: &GitRunInfo) -> eyre::Result<isize> {
+pub fn undo(
+    effects: &Effects,
+    git_run_info: &GitRunInfo,
+    interactive: bool,
+) -> eyre::Result<isize> {
     let repo = Repo::from_current_dir()?;
+    let references_snapshot = repo.get_references_snapshot()?;
     let conn = repo.get_db_conn()?;
-    let merge_base_db = MergeBaseDb::new(&conn)?;
     let mut event_log_db = EventLogDb::new(&conn)?;
     let mut event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
+    let dag = {
+        // Don't let `event_cursor` leak from this scope, since we intend to
+        // determine a new event cursor below.
+        let event_cursor = event_replayer.make_default_cursor();
+
+        Dag::open_and_sync(
+            effects,
+            &repo,
+            &event_replayer,
+            event_cursor,
+            &references_snapshot,
+        )?
+    };
 
     let event_cursor = {
-        let result = with_siv(effects, |effects, siv| {
-            select_past_event(siv, &effects, &repo, &merge_base_db, &mut event_replayer)
-        })?;
-        match result {
-            Some(event_cursor) => event_cursor,
-            None => return Ok(0),
+        if interactive {
+            let result = with_siv(effects, |effects, siv| {
+                select_past_event(siv, &effects, &repo, &dag, &mut event_replayer)
+            })?;
+            match result {
+                Some(event_cursor) => event_cursor,
+                None => return Ok(0),
+            }
+        } else {
+            event_replayer.advance_cursor_by_transaction(event_replayer.make_default_cursor(), -1)
         }
     };
 
@@ -826,19 +868,19 @@ pub mod testing {
 
     use cursive::{CursiveRunnable, CursiveRunner};
 
-    use crate::core::eventlog::{EventCursor, EventLogDb, EventReplayer};
-    use crate::core::mergebase::MergeBaseDb;
-    use crate::git::{GitRunInfo, Repo};
-    use crate::tui::Effects;
+    use lib::core::dag::Dag;
+    use lib::core::effects::Effects;
+    use lib::core::eventlog::{EventCursor, EventLogDb, EventReplayer};
+    use lib::git::{GitRunInfo, Repo};
 
     pub fn select_past_event(
         siv: CursiveRunner<CursiveRunnable>,
         effects: &Effects,
         repo: &Repo,
-        merge_base_db: &MergeBaseDb,
+        dag: &Dag,
         event_replayer: &mut EventReplayer,
     ) -> eyre::Result<Option<EventCursor>> {
-        super::select_past_event(siv, effects, repo, merge_base_db, event_replayer)
+        super::select_past_event(siv, effects, repo, dag, event_replayer)
     }
 
     pub fn undo_events(
@@ -866,7 +908,7 @@ pub mod testing {
 mod tests {
     use super::*;
 
-    use crate::core::eventlog::testing::make_dummy_transaction_id;
+    use lib::core::eventlog::testing::make_dummy_transaction_id;
 
     #[test]
     fn test_optimize_inverse_events() -> eyre::Result<()> {
